package org.telosys.tools.generator;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import org.telosys.tools.commons.TelosysToolsLogger;
import org.telosys.tools.commons.Variable;
import org.telosys.tools.generator.config.IGeneratorConfig;
import org.telosys.tools.generator.context.EntityContext;
import org.telosys.tools.generator.context.JavaBeanClass;
import org.telosys.tools.generator.context.JavaClass;
import org.telosys.tools.generator.target.EntityTarget;
import org.telosys.tools.repository.model.Entity;
import org.telosys.tools.repository.model.RepositoryModel;


public class GenerationManager 
{
	/**
	 * Generates a target using the given RepositoryModel and generator configuration
	 * 
	 * @param target 
	 * @param repositoryModel
	 * @param generatorConfig
	 * @param logger
	 * @throws GeneratorException
	 */
	public void generateTarget( EntityTarget target, RepositoryModel repositoryModel, IGeneratorConfig generatorConfig, 
			TelosysToolsLogger logger ) throws GeneratorException
	{
		String outputFileName = target.getFullFileName( generatorConfig.getProjectLocation() );
		
		InputStream is = null ;

		logger.info("Generation in progress : target = " + target.getTargetName() + " / entity = " + target.getEntityName() );
		is = generate(target.getEntityName(), target.getTemplate(), target.getJavaClassType(), 
						repositoryModel, generatorConfig, logger);
		logger.info("Generation done.");

		logger.info("Saving target file : " + outputFileName );
		saveInFile(is, outputFileName );
		logger.info("Target file saved." );
	}
	
	/**
	 * Saves the given InputStream in a file. The parent folders are created if necessary.
	 * @param is
	 * @param fileName
	 * @throws IOException
	 * @throws GeneratorException
	 */
	public void saveInFile(InputStream is, String fileName) throws GeneratorException
	{
		File f = new File(fileName);
		
		checkTargetFile(f, true);
		
		try {
			OutputStream out = new FileOutputStream(f);
			byte buf[] = new byte[1024];
			int len;
			while ((len = is.read(buf)) > 0) {
				out.write(buf, 0, len);
			}
			out.close();
			is.close();
		} catch (FileNotFoundException e) {
			throw new GeneratorException("Cannot save file "+fileName, e);
		} catch (IOException e) {
			throw new GeneratorException("Cannot save file "+fileName, e);
		}
	}
	
	protected void checkTargetFile(String sFile, boolean bCreateDir) throws GeneratorException 
	{
		if ( null == sFile ) {
			throw new GeneratorException("Target file is null !");
		}

		File file = new File(sFile);
		checkTargetFile(file, bCreateDir);
	}
	
	protected void checkTargetFile(File file, boolean bCreateDir) throws GeneratorException 
	{
		if ( null == file ) {
			throw new GeneratorException("Target file is null !");
		}
		
		if ( file.exists() )
		{
			if ( ! file.canWrite() )				
			{
				throw new GeneratorException("Cannot write on target file '"+ file.toString() + "' !");
			}
		}
		else
		{
			File parent = file.getParentFile();
			if ( ! parent.exists() )
			{
				if ( bCreateDir == false )
				{
					throw new GeneratorException("Target directory '"+ parent.toString() + "' not found !");
				}
				else
				{
					// Create the target file directory(ies)
					parent.mkdirs();				
				}
			}
		}
	}
	
	/**
	 * Generates an InputStream for the given target using the given RepositoryModel and generator configuration <br>
	 * NB : the result is only in memory (not saved in a file)
	 * 
	 * @param entityName
	 * @param templateFileName
	 * @param model
	 * @param generatorConfig
	 * @param logger
	 * @return
	 * @throws GeneratorException
	 */
	public InputStream generate( String entityName, String templateFileName, int javaClassType,
			RepositoryModel model, IGeneratorConfig generatorConfig, 
			TelosysToolsLogger logger ) throws GeneratorException
	{
		//--- Get the templates directory		
		String templatesDirectory = generatorConfig.getTemplatesFolderFullPath();		
		logger.log("Templates Directory : '" + templatesDirectory + "'");
		
		//--- Get the project variables	
		Variable[] projectVariables = generatorConfig.getProjectVariables();
		logger.log("Project variables count = " + ( projectVariables != null ? projectVariables.length : 0 ) );
		
		
		EntityContext entityContext = buildEntityContext(model, generatorConfig, entityName, logger );

		
		// Instanciate new generator
		//Generator generator = new Generator(templatesDirectory, templateFileName, projectVariables, logger);
		Generator generator = new Generator(templateFileName, generatorConfig, logger);

		generator.initEntityContext(entityContext, javaClassType);
		
		// Launch generation
		InputStream is = generator.generateStream();
		
		return is ;
	}
	
	protected EntityContext buildEntityContext(RepositoryModel model, IGeneratorConfig projectConfig, 
			String entityName, TelosysToolsLogger logger ) throws GeneratorException
    {
		if ( null == model )
		{
			throw new GeneratorException("RepositoryModel parameter is null");
		}
		if ( null == projectConfig )
		{
			throw new GeneratorException("ProjectConfig parameter is null");
		}
		if ( null == entityName )
		{
			throw new GeneratorException("Entity name parameter is null");
		}
		
		/*
		 * Retrieve the entity from the repository model 
		 *   
		 */
		Entity entity = model.getEntityByName(entityName);
		if ( null == entity ) 
		{
			throw new GeneratorException("Entity '" + entityName + "' not found in the repository");
		}
		if ( entityName.equals(entity.getName()) != true )
		{
			throw new GeneratorException("Repository corrupted : Entity name '" + entityName + "' != '" + entity.getName() +"'");
		}
		
		/*
		 * Entity generation context initialization
		 * It contains the 4 standard Java classes :
		 * . bean 
		 * . list
		 * . DAO
		 * . converter (mapper)
		 */
    	EntityContext ec = new EntityContext();

    	logger.log("EntityContext initialization " );

    	//--- Entity Bean Class (Value Object)
    	String beanClassName = entity.getBeanJavaClass();
    	String beanPackage = projectConfig.getVOPackage();
    	JavaBeanClass beanClass = new JavaBeanClass(entity, model, beanClassName, beanPackage, logger);    	
    	ec.setBeanClass(beanClass);
    	logger.log(". Bean Class : " + beanClass );
    	
    	//--- Bean List Class
    	String listClassName = entity.getListJavaClass();
    	String listPackage = projectConfig.getVOListPackage();
    	JavaClass listClass = new JavaClass(listClassName, listPackage);    	
    	ec.setListClass(listClass);
    	logger.log(". List Class : " + listClass );

    	//--- DAO Class
    	String daoClassName = entity.getDaoJavaClass();
    	String daoPackage   = projectConfig.getDaoPackage();
    	JavaClass daoClass = new JavaClass(daoClassName, daoPackage);    	
    	ec.setDaoClass(daoClass);
    	logger.log(". DAO Class  : " + daoClass );
    	
    	//--- Convertor Class
    	String converterClassName = entity.getConverterJavaClass();
    	String converterPackage   = projectConfig.getXmlMapperPackage();
    	JavaClass converterClass = new JavaClass(converterClassName, converterPackage);    	
    	ec.setConverterClass(converterClass);
    	logger.log(". Conv Class : " + converterClass );
    	
    	return ec;
    }

}
