package org.telosys.tools.generator.context;

import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.telosys.tools.commons.StrUtil;
import org.telosys.tools.commons.TelosysToolsLogger;
import org.telosys.tools.generator.GeneratorUtil;
import org.telosys.tools.generator.context.tools.AnnotationsBuilder;
import org.telosys.tools.repository.model.Column;
import org.telosys.tools.repository.model.Entity;
import org.telosys.tools.repository.model.Link;
import org.telosys.tools.repository.model.RepositoryModel;

/**
 * Specific Java Class for an entity Java Bean with Object-Relational Mapping (ORM) <br>
 * This class provides the standard Java class informations plus : <br>
 * . the attributes of the class <br>
 * . the imports required by attributes types <br>
 * . the database table where the object is stored <br>
 * . the mapping for each attribute<br>
 * 
 * @author Laurent GUERIN
 *
 */
public class JavaBeanClass extends JavaClass
{
	
	private final static Object[] VOID_ARRAY = new Object[0] ;
	
//	private final TelosysToolsLogger _logger;
	
	private List<String>       _imports     = null ; // The imports for this class ( list of "java.xx.Class" )
	private LinkedList<JavaBeanClassAttribute> _attributes  = null ; // The attributes for this class ( ALL ATTRIBUTES )
	
    private String     _sDatabaseTable   = null ; // Table name this class is mapped with
    private String     _sDatabaseCatalog = null ; // The table's catalog 
    private String     _sDatabaseSchema  = null ; // The table's schema 
    
	private LinkedList _keyAttributes  = null ; // The KEY attributes for this class
	private LinkedList _nonKeyAttributes  = null ; // The NON KEY attributes for this class
	private String     _sSqlKeyColumns = null ;
	private String     _sSqlNonKeyColumns = null ;
	
	//--- XML mapper infos
	private LinkedList _nonTextAttributes  = null ; // Standard attributes for this class ( not "long text" )
	private LinkedList _textAttributes  = null ; // Special "long text" attributes for this class

	//--- JPA specific
	private Set<String>       _importsJpa     = null ; // The imports JPA for this class ( list of "java.xx.Class" )
	private LinkedList<JavaBeanClassLink> _links  = null ; // The links for this class ( ALL ATTRIBUTES )
	private Entity _entite = null;
	
	//-----------------------------------------------------------------------------------------------
	/**
	 * Constructor based on Repository Entity
	 * @param entity
	 * @param model
	 * @param nameEntity
	 * @param sPackage
	 * @param logger
	 */
	public JavaBeanClass(final Entity entity, final RepositoryModel model, final String nameEntity, 
			final String sPackage, 
			final TelosysToolsLogger logger) 
	{
		super(entity.getBeanJavaClass(), sPackage);
		
		//this._logger = logger;

		if ( null == entity ) {
			throw new RuntimeException("Cannot create JavaClass : Entity is null");
		}
		
		this._entite = entity;
		this._sDatabaseTable = entity.getName();
		_sDatabaseCatalog = entity.getCatalog();
		_sDatabaseSchema  = entity.getSchema();
		
		/*
		 * Set all the attributes of the current Java class
		 *  
		 */
		for (Iterator iterator = entity.getColumnsCollection().iterator(); iterator.hasNext();) 
		{
			Column col = (Column) iterator.next();
			//JavaBeanClassAttribute jca = new JavaBeanClassAttribute(col, logger);
			JavaBeanClassAttribute jca = new JavaBeanClassAttribute(col);
			this.addAttribute(jca);
		}

		/*
		 * Set all the links of the current Java class
		 */
		for (Iterator iterator = entity.getLinksCollection().iterator(); iterator.hasNext();) 
		{
			Link link = (Link) iterator.next();
			
			// On va trouver le bean correspondant a ce lien dans le model
			Entity entityCible = model.getEntityByName(link.getTargetTableName());
			JavaBeanClassLink jcl = new JavaBeanClassLink(link, this._entite , entityCible, logger);
			
			//ajouter import specifique
			//JavaBeanClassImports jbci = new JavaBeanClassImports(_logger);
			JavaBeanClassImports jbci = new JavaBeanClassImports();
			jbci.declareType(jcl.getJavaTypeFull());
			this.addImportsJpa(jbci);
			this.addLink(jcl);
		}
		
		// import resolution
		this.endOfDefinition();
		
		// import JPA resolution
		this.processJpaSpecificImport();
	}
	
	//-----------------------------------------------------------------------------------------------
	/**
	 * Constructor based on a list of attributes
	 * @param sFullClassName
	 * @param attributes
	 */
	public JavaBeanClass( String sFullClassName, JavaBeanClassAttribute[] attributes ) 
	{
		super( sFullClassName );
		initAttributes( attributes ); 
	}
	
	//-----------------------------------------------------------------------------------------------
	/**
	 * Constructor based on a list of attributes
	 * @param sShortClassName
	 * @param sPackage
	 * @param attributes
	 */
	public JavaBeanClass( String sShortClassName, String sPackage, JavaBeanClassAttribute[] attributes ) 
	{
		super( sShortClassName, sPackage);
		initAttributes( attributes ); 
	}
	
	//-----------------------------------------------------------------------------------------------
	private void initAttributes(JavaBeanClassAttribute[] attributes) 
	{
		//--- Add each attribute 
		if ( attributes != null ) 
		{
			for ( JavaBeanClassAttribute attribute : attributes )
			{
				addAttribute(attribute);
			}
			endOfDefinition(); // close the class definition (prepares imports list)
		}
		
	}
	//-----------------------------------------------------------------------------------------------
	private void addLink(JavaBeanClassLink jcl) {
		if ( _links == null )
		{
			_links = new LinkedList<JavaBeanClassLink>();
		}
		_links.add(jcl);
	}

	
//	/**
//	 * a quoi sert ce constructeur ????
//	 * @deprecated
//	 */
//	public JavaBeanClass(final Entity entite, final RepositoryModel model, final String nameEntity, 
//			final String sPackage, 
//			final JavaBeanClassImports imports,
//			final TelosysToolsLogger logger) 
//	{
//		//--- Definition with automatic imports determination
//		this(entite, model, nameEntity, sPackage, logger );
//
//		//--- Force the imports list
//		_imports = new LinkedList();
//		_imports.addAll( imports.getList() );
//		
//	}
	
	/**
	 * Returns the JPA annotations without left margin
	 * Usage : $x.jpaAnnotations 
	 * @return
	 */
	public String getJpaAnnotations()
    {
		return jpaAnnotations(0);
    }
	/**
	 * Returns the JPA annotations without left margin 
	 * Usage : $x.jpaAnnotations() 
	 * @return
	 */
	public String jpaAnnotations()
    {
		return jpaAnnotations(0);
    }
	/**
	 * Returns the JPA annotations with the given left margin 
	 * Usage : $x.jpaAnnotations(4) 
	 * @param iLeftMargin
	 * @return
	 */
	public String jpaAnnotations(int iLeftMargin)
    {
//		StringBuilder jpaAnnot = new StringBuilder();
//		jpaAnnot.append("@Entity\n");
//		jpaAnnot.append("@Table(name=\"");
//		jpaAnnot.append(this._sDatabaseTable);
//		jpaAnnot.append("\")");
//        return jpaAnnot.toString();
		
		AnnotationsBuilder b = new AnnotationsBuilder(iLeftMargin);
//		b.resetAnnotations(iLeftMargin) ;
		
		b.addLine("@Entity");
		
		String s = "@Table(name=\"" + _sDatabaseTable + "\"" ;
		if ( ! StrUtil.nullOrVoid( _sDatabaseSchema ) ) {
			s = s + ", schema=\"" + _sDatabaseSchema + "\"" ; 
		}
		if ( ! StrUtil.nullOrVoid( _sDatabaseCatalog ) ) {
			s = s + ", catalog=\"" + _sDatabaseCatalog + "\"" ; 
		}
		s = s + " )" ;

		b.addLine(s);
		
		return b.getAnnotations();
    }
	
	/**
	 * Returns an array of imports
	 * @return
	 */
	public Object[] getImports() 
	{
		if ( _imports != null )
		{
			Object[] array = _imports.toArray();
			return array ;
		}
		return VOID_ARRAY ;
	}

	/**
	 * Returns an array of imports JPA
	 * @return
	 */
	public Object[] getImportsJpa() 
	{
		if ( _importsJpa != null )
		{
			Object[] array = _importsJpa.toArray();
			return array ;
		}
		return VOID_ARRAY ;
	}

	/**
	 * Returns all the attributes defined for this class
	 * @return
	 */
	public Object[] getAttributes() 
	{
		if ( _attributes != null )
		{
			Object[] array = _attributes.toArray();
			return array ;
		}
		return VOID_ARRAY ;
	}

	/**
	 * Returns all the link defined for this class
	 * @return
	 */
	public Object[] getLinks() 
	{
		if ( _links != null )
		{
			Object[] array = _links.toArray();
			return array ;
		}
		return VOID_ARRAY ;
	}

	/**
	 * Returns the list of "key" attributes ( PRIMARY KEY ELEMENTS )
	 * @return
	 */
	public Object[] getKeyAttributes() 
	{
		if ( _attributes != null )
		{
			if ( _keyAttributes == null ) // list not yet built
			{
				_keyAttributes = buildAttributesList ( true ); 
			}
			Object[] array = _keyAttributes.toArray();
			return array ;
		}
		return VOID_ARRAY ;
	}

	/**
	 * Returns the list of "non key" attributes ( NON PRIMARY KEY ELEMENTS )
	 * @return
	 */
	public Object[] getNonKeyAttributes() 
	{
		if ( _attributes != null )
		{
			if ( _nonKeyAttributes == null ) // list not yet built
			{
				_nonKeyAttributes = buildAttributesList ( false ); 
			}
			Object[] array = _nonKeyAttributes.toArray();
			return array ;
		}
		return VOID_ARRAY ;
	}

	public String getSqlTable() 
	{
		return _sDatabaseTable ;
	}
	
	public String getDatabaseCatalog() 
	{
		return _sDatabaseCatalog ;
	}
	
	public String getDatabaseSchema() 
	{
		return _sDatabaseSchema ;
	}
	
	/**
	 * Returns a string containing all the colums of the Primary Key separated by a comma <br>
	 * Example : '"code", "type"' 
	 * @return
	 */
	public String getSqlKeyColumns() 
	{
		if ( _sSqlKeyColumns == null ) // list not yet built
		{
			_sSqlKeyColumns = buildDbColumnsList( true ); 
		}
		return _sSqlKeyColumns ;
	}
	
	/**
	 * Returns a string containing all the colums that are not in the Primary Key separated by a comma <br>
	 * Example : '"first_name", "age", "email"' 
	 * @return
	 */
	public String getSqlNonKeyColumns() 
	{
		if ( _sSqlNonKeyColumns == null ) // list not yet built
		{
			_sSqlNonKeyColumns = buildDbColumnsList( false ); 
		}
		return _sSqlNonKeyColumns ;
	}

    /**
     * Returns the Java line instruction for the toString() method
     * @return
     */
    public String getToStringInstruction()
    {
    	if ( _attributes != null )
    	{
    		int n = _attributes.size();
    		if ( n > 1 )
    		{
                StringBuffer sb = new StringBuffer();
            	for ( int i = 0 ; i < n ; i++ )        		
            	{
            		JavaBeanClassAttribute attribute = (JavaBeanClassAttribute) _attributes.get(i);
                    if ( i > 0 ) // if it's not the first one
                    {
                        sb.append( " + \"|\" + " ) ;
                    }        		
                    sb.append( attribute.getName() ) ;
            	}
            	return sb.toString(); // example : 'aaa + "|" + bbb + "|" + ccc'
    		}
    		else 
    		{
    			// Single attribute => no automatic conversion to String
    			JavaBeanClassAttribute attribute = (JavaBeanClassAttribute) _attributes.get(0);
    			String sFullType = attribute.getFullType();
    			if ( sFullType != null ) {
    				if ( sFullType.startsWith("java.") ) {
    					//--- Java object
    					if ( sFullType.equals("java.lang.String") ) {
        					return attribute.getName() ;
    					}
    					else {
        					return attribute.getName() + ".toString()";
    					}
    				}
    				else {
    					//--- Primitive type
    					return "\"\" + " + attribute.getName() ;
    				}
    			}
    		}
    	}
    	return "\"Class " + getName() + " (no attributes) \"" ;
    }

    private boolean typeUsedInToString( String sType )
    {
    	if ( null == sType ) return false ;
    	String s = sType.trim() ;
    	if ( s.endsWith("]") ) return false ; // Array
    	if ( s.endsWith("Blob") ) return false ; 
    	if ( s.endsWith("Clob") ) return false ; 
    	return true ;
    }
    public String toStringMethodCodeLines( int iLeftMargin )
    {
    	String leftMargin = GeneratorUtil.blanks(iLeftMargin);
    	
    	if ( _attributes != null )
    	{
    		int n = _attributes.size();
    		if ( n > 0 )
    		{
    			int count = 0 ;
                StringBuffer sb = new StringBuffer();
    			sb.append(leftMargin);
    			sb.append("StringBuffer sb = new StringBuffer(); \n");
            	for ( int i = 0 ; i < n ; i++ )        		
            	{
            		JavaBeanClassAttribute attribute = (JavaBeanClassAttribute) _attributes.get(i);
            		if ( typeUsedInToString( attribute.getType() ) )
            		{
                        if ( count > 0 ) // if it's not the first one
                        {
                			sb.append(leftMargin);
                			sb.append("sb.append(\"|\"); \n");
                        }        		
            			sb.append(leftMargin);
            			sb.append("sb.append(");
            			sb.append(attribute.getName());
            			sb.append("); \n");
            			count++ ;
            		}
            	}
    			sb.append(leftMargin);
    			sb.append("return sb.toString();"); // Last line => No EOL
    			if ( count > 0 ) {
                    return sb.toString() ;
    			}
    		}
    	}
    	return leftMargin + "return \"Instance of class " + getName() + "\" ;" ;
    }

	/**
	 * Returns an array containing all the "NON long text" attributes of this class (standard attributes)<br>
	 * ( returns a void array if no attributes )
	 * @return
	 */
	public Object[] getNonTextAttributes() 
	{
		if ( _nonTextAttributes == null ) // list not yet built
		{
			_nonTextAttributes = buildTextAttributesList ( false ); // NOT LONG TEXT
		}
		return toArray(_nonTextAttributes) ;
	}

	/**
	 * Returns an array containing all the "long text" attributes of this class<br>
	 * ( returns a void array if no "long text" attributes )
	 * @return
	 */
	public Object[] getTextAttributes() 
	{
		if ( _textAttributes == null ) // list not yet built
		{
			_textAttributes = buildTextAttributesList ( true ); // Special "LONG TEXT"
		}
		return toArray(_textAttributes) ;
	}

	/**
	 * Returns true if this class has at least one "long text" attribute
	 * @return
	 */
	public boolean getHasTextAttribute() 
	{
    	if ( _attributes != null )
    	{
    		int n = _attributes.size();
        	for ( int i = 0 ; i < n ; i++ )        		
        	{
        		JavaBeanClassAttribute attribute = (JavaBeanClassAttribute) _attributes.get(i);
                if ( attribute.isLongText() ) 
                {
                	return true ;
                }
        	}
    	}
    	return false ;
	}

//	/**
//	 * Returns true if this class has an "autoincremented" Primary Key <br>
//	 * A Primary Key with only one attribute that is "autoincremented"
//	 * @return
//	 */
//	public boolean getHasAutoincrementedKey() 
//	{
//    	if ( _attributes != null )
//    	{
//    		int n = 0 ;
//        	for ( JavaBeanClassAttribute attribute : _attributes )        		
//        	{
//                if ( attribute.isKeyElement() ) {
//                    if ( attribute.isAutoIncremented() ) {
//                    	n++;
//                    }
//                }
//        	}
//    	}
//    	return false ;
//	}

	public JavaBeanClassAttribute getAutoincrementedKeyAttribute() 
	{
		Object [] keyAttributes = getKeyAttributes();
    	if ( keyAttributes != null )
    	{
    		if ( keyAttributes.length == 1 ) 
    		{
    			// Only one attribute in the PK
    			JavaBeanClassAttribute attribute = (JavaBeanClassAttribute) keyAttributes[0];
    			if ( attribute != null ) {
                    if ( attribute.isAutoIncremented() ) 
                    {
                    	// This unique PK field is auto-incremented => return it
                    	return attribute ; 
                    }
    			}
    		}
    	}
    	return null ;
	}

	/* (non-Javadoc)
	 * Same as getName() 
	 * @see java.lang.Object#toString()
	 */
	public String toString()
	{
		return super.toString() ;
	}
	
	
	// -------------------------------------------------------------------------------------------------
	
	/**
	 * Add an attribute for this class
	 * @param attribute
	 */
    private void addAttribute(JavaBeanClassAttribute attribute) 
	{
		if ( _attributes == null )
		{
			_attributes = new LinkedList<JavaBeanClassAttribute>();
		}
		_attributes.add(attribute);
		//PluginLogger.log("addAttribute(attribute)");
	}
	
	/**
	 * Init the imports list
	 * @param imports
	 */
    private void setImports(JavaBeanClassImports imports) 
	{
		if ( imports != null )
		{
			// Reset ALL => create a new list
			_imports = new LinkedList<String>();
			_imports.addAll( imports.getList() );
		}
	}
    

	/**
	 * Init the Jpa imports list
	 * @param imports
	 */
    private void addImportsJpa(JavaBeanClassImports imports) 
	{
		if ( imports != null )
		{
			// Reset ALL => create a new list
			if (_importsJpa == null) {
				_importsJpa = new LinkedHashSet<String>();
			}
			
			_importsJpa.addAll( imports.getList() );
		}
	}
    
	private LinkedList<JavaBeanClassAttribute> buildAttributesList ( boolean bKeyAttribute ) 
	{
    	if ( _attributes != null )
    	{
			LinkedList<JavaBeanClassAttribute> list = new LinkedList<JavaBeanClassAttribute>();
    		int n = _attributes.size();
        	for ( int i = 0 ; i < n ; i++ )        		
        	{
        		//JavaBeanClassAttribute attribute = (JavaBeanClassAttribute) _attributes.get(i);
        		JavaBeanClassAttribute attribute = _attributes.get(i);
                if ( attribute.isKeyElement() == bKeyAttribute ) 
                {
                	list.add(attribute);
                }        		
        	}
    		return list ;
    	}
    	return null ;
	}

	/**
	 * This method close the definition of the class (when all the attributes have been added) <br>
	 * 
	 * It determines if there is import types collision ( eg "java.util.Date" with "java.sql.Date" ) <br>
	 * and managed the imports list and attributes declarations types to avoid imports error
	 *  
	 */
	private void endOfDefinition() 
	{
		if ( _attributes == null ) return ;
		
		//--- Define the imports required for this class 
		//PluginLogger.log("Define imports ... " );
		//JavaBeanClassImports javaImports = new JavaBeanClassImports(this._logger);
		JavaBeanClassImports javaImports = new JavaBeanClassImports();
		
		Iterator iter = _attributes.iterator();
		while ( iter.hasNext() )
		{
			JavaBeanClassAttribute attr = (JavaBeanClassAttribute) iter.next();
			javaImports.declareType( attr.getFullType() ); // register the type to import if necessary
		}

		//--- Extract potential collided types ( and retrieve the list of collided full types )
		LinkedList collidedTypes = javaImports.extractDuplicatedShortNames();

		//--- Set imports list for the current class
		this.setImports(javaImports);

		//--- If there's collided types => Chech each attribute type 
		if ( collidedTypes != null )
		{
			//--- Some collided types have been extracted from imports 
			iter = _attributes.iterator();
			while ( iter.hasNext() )
			{
				JavaBeanClassAttribute attr = (JavaBeanClassAttribute) iter.next();
				String sFullType = attr.getFullType();
				if ( collidedTypes.contains( sFullType ) ) // if this attribute is impacted
				{
					//--- force this attributes to use its "full type" for variable declaration
					attr.forceType ( sFullType );
				}
			}
		}
	}

	/**
	 * This method process the jpa specific imports <br>
	 */
	private void processJpaSpecificImport() {
		//JavaBeanClassImports jpaImports = new JavaBeanClassImports(this._logger);
		JavaBeanClassImports jpaImports = new JavaBeanClassImports();

		// TODO a afiner
		jpaImports.declareType("javax.persistence.*");
		
		this.addImportsJpa(jpaImports);
		
		/*
		jpaImports.declareType("javax.persistence.Entity");
		jpaImports.declareType("javax.persistence.Table");
		jpaImports.declareType("javax.persistence.Id");
		
		jpaImports.declareType("javax.persistence.UniqueConstraint");
		jpaImports.declareType("javax.persistence.EmbeddedId");
		jpaImports.declareType("javax.persistence.Embeddable");
		jpaImports.declareType("javax.persistence.AttributeOverride");
		jpaImports.declareType("javax.persistence.AttributeOverrides");

		jpaImports.declareType("javax.persistence.OneToOne");
		jpaImports.declareType("javax.persistence.ManyToMany");
		jpaImports.declareType("javax.persistence.ManyToOne");
		jpaImports.declareType("javax.persistence.OneToMany");

		jpaImports.declareType("javax.persistence.GeneratedValue");
		jpaImports.declareType("javax.persistence.GenerationType");
		jpaImports.declareType("javax.persistence.SequenceGenerator");
		jpaImports.declareType("javax.persistence.TableGenerator");
		*/
	}

	private String buildDbColumnsList ( boolean bKeyAttribute ) 
	{
    	if ( _attributes != null )
    	{
            StringBuffer sb = new StringBuffer(60);
            int iCount = 0 ;
    		int n = _attributes.size();
        	for ( int i = 0 ; i < n ; i++ )        		
        	{
        		JavaBeanClassAttribute attribute = (JavaBeanClassAttribute) _attributes.get(i);
                if ( attribute.isKeyElement() == bKeyAttribute ) 
                {
                	if ( iCount > 0 ) // Not the first one
                	{
                        sb.append( ", " ) ;
                	}
                    sb.append( "\"" + attribute.getDatabaseName().trim() + "\"" ) ;
                    iCount++ ;
                }        		
        	}
    		return sb.toString() ;
    	}
    	return "" ;
	}

	/**
	 * "Text" or "non Text" attributes
	 * @param bLongText
	 * @return
	 */
	private LinkedList<JavaBeanClassAttribute> buildTextAttributesList ( boolean bLongText ) 
	{
    	if ( _attributes != null )
    	{
			LinkedList<JavaBeanClassAttribute> list = new LinkedList<JavaBeanClassAttribute>();
    		int n = _attributes.size();
        	for ( int i = 0 ; i < n ; i++ )        		
        	{
        		//JavaBeanClassAttribute attribute = (JavaBeanClassAttribute) _attributes.get(i);
        		JavaBeanClassAttribute attribute = _attributes.get(i);
                if ( attribute.isLongText() == bLongText ) 
                {
                	list.add(attribute);
                }        		
        	}
    		return list ;
    	}
    	return null ;
	}

	private Object[] toArray(LinkedList list)
	{
		if ( list != null )
		{
			return list.toArray();
		}
		return VOID_ARRAY ;
	}

}
