package org.telosys.tools.generator.target;

import org.telosys.tools.commons.StrUtil;
import org.telosys.tools.commons.Variable;
import org.telosys.tools.commons.VariablesManager;
import org.telosys.tools.commons.config.ConfigDefaults;
import org.telosys.tools.generator.GeneratorConst;
import org.telosys.tools.repository.model.Entity;

/**
 * A target specialized for a specific entity<br>
 * 
 * Concrete target built from a GenericTarget.
 * This concrete target doesn't contain variables in the file or folder name.
 * 
 * @author Laurent Guerin
 *
 */
public class EntityTarget 
{

	private String  _sName = null ;

	private String  _sFile = null ;
	
	private String  _sFolder = null ;

	private String  _sTemplate = null ;

	private boolean _bVariablesReplacement = false ;

	
	private String  _sEntityName = null ;
	
	private int     _iJavaClassType = GeneratorConst.JAVA_CLASS_TYPE_UNDEFINED ;

	/**
	 * Copy constructor
	 * @param t
	 */
	//public EntityTarget( GenericTarget genericTarget, Entity entity, VariablesManager varmanager ) 
	public EntityTarget( GenericTarget genericTarget, Entity entity, Variable[] variables ) 
	{
		super();
		
		//--- Generic target informations
		_sName = genericTarget.getName();
		_sFile = genericTarget.getFile();
		_sFolder = genericTarget.getFolder();
		_sTemplate = genericTarget.getTemplate();
		
		//--- Specialization for the given entity
		_sEntityName = entity.getName() ;
		
		//--- Replace the "$" variables in _sFile and _sFolder
		VariablesManager variablesManager = new VariablesManager( variables );
		replaceVariables(entity.getBeanJavaClass(), variablesManager);
		
		//--- For special generic targets with a file name defined in the repository ( Bean, List, DAO, Converter )
		_iJavaClassType = GeneratorConst.JAVA_CLASS_TYPE_UNDEFINED ;
		
		if ( _sFile.equals(GenericTarget.JAVA_CLASS_FOR_BEAN) ) {
			_sFile = entity.getBeanJavaClass() + ".java" ;
			_iJavaClassType = GeneratorConst.JAVA_CLASS_TYPE_BEAN ;
		}
		
		if ( _sFile.equals(GenericTarget.JAVA_CLASS_FOR_LIST) ) {
			_sFile = entity.getListJavaClass()+ ".java" ;
			_iJavaClassType = GeneratorConst.JAVA_CLASS_TYPE_LIST ;
		}
		
		if ( _sFile.equals(GenericTarget.JAVA_CLASS_FOR_DAO) ) {
			_sFile = entity.getDaoJavaClass()+ ".java" ;
			_iJavaClassType = GeneratorConst.JAVA_CLASS_TYPE_DAO ;
		}
		
		if ( _sFile.equals(GenericTarget.JAVA_CLASS_FOR_CONV) ) {
			_sFile = entity.getConverterJavaClass()+ ".java" ;
			_iJavaClassType = GeneratorConst.JAVA_CLASS_TYPE_CONVERTER ;
		}
	}
	
	//-----------------------------------------------------------------------
	/**
	 * Apply the bean class name to the current generic target<br>
	 * This method replaces the symbolic variables BEANNAME, BEANNAME_UC and BEANNAME_LC
	 * by the given bean class name
	 * And any project variable by its value
	 * @param sBeanClass
	 */
	private void replaceVariables( String sBeanClass, VariablesManager varmanager )
	{
		if ( _bVariablesReplacement != true )
		{
			//--- Replace the generic name "${BEANNAME}" if any
			_sFile   = replace(_sFile,   ConfigDefaults.BEANNAME, sBeanClass);
			_sFolder = replace(_sFolder, ConfigDefaults.BEANNAME, sBeanClass);

			//--- Replace the global project variables if any
			_sFile   = varmanager.replaceVariables(_sFile);
			_sFolder = varmanager.replaceVariables(_sFolder);
			
			_bVariablesReplacement = true ;
		}
	}
	//-----------------------------------------------------------------------
    private String replace(String sOriginal, String sSymbolicVar, String sValue) 
    {
    	String s   = "${" + sSymbolicVar + "}" ;
    	String sUC = "${" + sSymbolicVar + "_UC}" ;
    	String sLC = "${" + sSymbolicVar + "_LC}" ;
    	
		if ( sOriginal.indexOf(s) >= 0 )
		{
			return StrUtil.replaceVar(sOriginal, s, sValue);
		}
		else if ( sOriginal.indexOf(sUC) >= 0 )
		{
			return StrUtil.replaceVar(sOriginal, sUC, sValue.toUpperCase());
		}
		else if ( sOriginal.indexOf(sLC) >= 0 )
		{
			return StrUtil.replaceVar(sOriginal, sLC, sValue.toLowerCase());
		}
		return sOriginal ;
    }
    
	//-----------------------------------------------------------------------
	/**
	 * Returns the entity name 
	 * @return
	 */
	public String getEntityName()
	{
		return _sEntityName ;
	}
	
	//-----------------------------------------------------------------------
	/**
	 * Returns the target name ( the text displayed on the screen )
	 * @return
	 */
	public String getTargetName()
	{
		return _sName ;
	}
	
	/**
	 * Returns the target file name ( file to be generated )
	 * @return
	 */
	public String getFile()
	{
		return _sFile ;
	}
	
	/**
	 * Returns the folder where to generate the file
	 * @return
	 */
	public String getFolder()
	{
		return _sFolder ;
	}
	
	/**
	 * Returns the java class type ( Bean, List, DAO, Converter or Undefined ).
	 * See constants 
	 * @return
	 */
	public int getJavaClassType()
	{
		return _iJavaClassType ;
	}
	
	/**
	 * Returns the full path of the target file in the project<br>
	 * ie : "src/org/demo/screen/employee/EmployeeData.java"
	 * @return
	 */
	public String getFullPathInProject()
	{
		String s = null ;
		if ( _sFolder.endsWith("/") || _sFolder.endsWith("\\") )
		{
			s = _sFolder + _sFile ;
		}
		else
		{
			s = _sFolder + "/" + _sFile ;
		}
		if ( s.startsWith("/") || s.startsWith("\\") )
		{
			return s.substring(1);
		}
		return s ;
	}

	/**
	 * Returns the absolute full path of the target in the file system <br>
	 * ie : "C:/tmp/project/src/org/demo/screen/employee/EmployeeData.java"
	 * @param sProjectLocation the project location ( ie "C:/tmp/project" )
	 * @return
	 */
	public String getFullFileName(String sProjectLocation)
	{
		String s = getFullPathInProject();
		if ( sProjectLocation != null )
		{
			if ( sProjectLocation.endsWith("/") || sProjectLocation.endsWith("\\") )
			{
				return sProjectLocation + s ;
			}
			else
			{
				return sProjectLocation + "/" + s ;
			}
		}
		return "/" + s ;
	}
	
	/**
	 * Returns the template 
	 * @return
	 */
	public String getTemplate()
	{
		return _sTemplate ;
	}	
	
	public String toString()
	{
		return _sName + " : '" + _sFile  + "' in '" + _sFolder + "' ( " + _sTemplate + " )" ;
	}
}
